/**
 * @license
 * Copyright 2023 Google LLC
 * SPDX-License-Identifier: Apache-2.0
 */
var __addDisposableResource = (this && this.__addDisposableResource) || function (env, value, async) {
    if (value !== null && value !== void 0) {
        if (typeof value !== "object" && typeof value !== "function") throw new TypeError("Object expected.");
        var dispose;
        if (async) {
            if (!Symbol.asyncDispose) throw new TypeError("Symbol.asyncDispose is not defined.");
            dispose = value[Symbol.asyncDispose];
        }
        if (dispose === void 0) {
            if (!Symbol.dispose) throw new TypeError("Symbol.dispose is not defined.");
            dispose = value[Symbol.dispose];
        }
        if (typeof dispose !== "function") throw new TypeError("Object not disposable.");
        env.stack.push({ value: value, dispose: dispose, async: async });
    }
    else if (async) {
        env.stack.push({ async: true });
    }
    return value;
};
var __disposeResources = (this && this.__disposeResources) || (function (SuppressedError) {
    return function (env) {
        function fail(e) {
            env.error = env.hasError ? new SuppressedError(e, env.error, "An error was suppressed during disposal.") : e;
            env.hasError = true;
        }
        function next() {
            while (env.stack.length) {
                var rec = env.stack.pop();
                try {
                    var result = rec.dispose && rec.dispose.call(rec.value);
                    if (rec.async) return Promise.resolve(result).then(next, function(e) { fail(e); return next(); });
                }
                catch (e) {
                    fail(e);
                }
            }
            if (env.hasError) throw env.error;
        }
        return next();
    };
})(typeof SuppressedError === "function" ? SuppressedError : function (error, suppressed, message) {
    var e = new Error(message);
    return e.name = "SuppressedError", e.error = error, e.suppressed = suppressed, e;
});
import { inspect } from 'util';
import { scriptReferenceToString, } from '../../config.js';
import { DefaultLogger, labelForScript } from '../default-logger.js';
import { DEBUG } from '../logger.js';
import { StackMap } from './stack-map.js';
/**
 * State that the run tracker cares about for a currently running script.
 */
class ScriptState {
    constructor(scriptReference, service, isRootScript) {
        this.outputBuffer = [];
        this.scriptReference = scriptReference;
        this.service = service;
        this.isRootScript = isRootScript;
    }
    bufferOutput(output) {
        if (output.data.length === 0) {
            return;
        }
        this.outputBuffer.push({
            stream: output.stream,
            data: output.data,
        });
    }
    replayAndEmptyBuffer() {
        for (const output of this.outputBuffer) {
            if (output.stream === 'stdout') {
                process.stdout.write(output.data);
            }
            else {
                process.stderr.write(output.data);
            }
        }
        this.outputBuffer.length = 0;
    }
    get hasBufferedOutput() {
        return this.outputBuffer.length > 0;
    }
}
export const noChange = Symbol('nochange');
export const nothing = Symbol('nothing');
/**
 * Tracks the state of a run, and produces a single line of status text.
 *
 * A QuietLogger usually just has one of these, but in --watch mode we use
 * one per iteration.
 */
export class QuietRunLogger {
    constructor(rootPackage, statusLineWriter, defaultLogger) {
        /**
         * Currently running scripts, or failed scripts that we're about to
         * report on. A script is added to this when it starts, and removed
         * only when it successfully exits.
         *
         * Keyed by this._getKey
         */
        this._running = new StackMap();
        /** The number of commands we've started. */
        this._ran = 0;
        /**
         * The number of scripts with commands that were either fresh or whose output
         * we restored.
         */
        this._skipped = 0;
        this._encounteredFailures = false;
        this._servicesRunning = 0;
        this._servicesStarted = 0;
        this._servicesPersistedFromPreviousRun = 0;
        this._analysisInfo = undefined;
        this._finishedScriptsWithCommands = new Set();
        this._statusLineState = 'initial';
        this._startTime = Date.now();
        /**
         * Sometimes a script will fail multiple times, but we only want to report
         * about the first failure for it that we find. Sometimes a script will
         * even end up with multiple failures for the same reason, like multiple
         * non-zero-exit-code events. This looks to be coming at the NodeJS level
         * or above, and so we just cope.
         */
        this._scriptsWithAlreadyReportedErrors = new Set();
        this._rootPackage = rootPackage;
        this._statusLineWriter = statusLineWriter;
        this._defaultLogger = defaultLogger ?? new DefaultLogger(rootPackage);
    }
    /**
     * Used to make a new instance, keeping info about persistent services
     * that continue from the previous run to the next.
     */
    makeInstanceForNextWatchRun() {
        // Reuse the default logger, a minor savings.
        const instance = new QuietRunLogger(this._rootPackage, this._statusLineWriter, this._defaultLogger);
        // Persistent services stay running between runs, so pass along what we
        // know.
        for (const [key, state] of this._running) {
            if (state.service) {
                instance._servicesPersistedFromPreviousRun++;
                instance._servicesRunning++;
                instance._running.set(key, state);
                instance._markScriptAsFinished(state.scriptReference);
            }
        }
        return instance;
    }
    /**
     * Takes an Event and updates our summary stats for this run.
     *
     * If it returns undefined, the previous message is still good. If it returns
     * null, the message should be cleared entirely.
     */
    getUpdatedMessageAfterEvent(event) {
        switch (event.type) {
            case 'success': {
                return this._handleSuccess(event);
            }
            case 'failure': {
                return this._handleFailure(event);
            }
            case 'info': {
                return this._handleInfo(event);
            }
            case 'output': {
                return this._handleOutput(event);
            }
            default: {
                const never = event;
                throw new Error(`Unknown event type: ${JSON.stringify(never)}`);
            }
        }
    }
    /**
     * Should be called once, at the end of a run.
     */
    printSummary() {
        this._statusLineState = 'done';
        let scriptsStillRunning = false;
        for (const state of this._running.values()) {
            if (state.service) {
                continue;
            }
            scriptsStillRunning = true;
        }
        if (this._encounteredFailures || scriptsStillRunning) {
            this._printFailureSummary();
            return;
        }
        this._printSuccessSummary();
    }
    _printFailureSummary() {
        for (const [, state] of this._running) {
            const label = labelForScript(this._rootPackage, state.scriptReference);
            const key = scriptReferenceToString(state.scriptReference);
            if (this._scriptsWithAlreadyReportedErrors.has(key) || state.service) {
                continue;
            }
            // We expected this to finish running, but it didn't. We haven't seen
            // this in our testing, so it's as much of a test of our own code as
            // anything else.
            process.stderr.write(`\n❌ [${label}] did not exit successfully.`);
            this._reportOutputForFailingScript(state.scriptReference);
        }
        if (this._scriptsWithAlreadyReportedErrors.size > 0) {
            const s = this._scriptsWithAlreadyReportedErrors.size === 1 ? '' : 's';
            process.stderr.write(`❌ ${this._scriptsWithAlreadyReportedErrors.size.toLocaleString()} script${s} failed.\n`);
        }
        else {
            process.stderr.write(`❌ Failed.\n`);
        }
    }
    _printSuccessSummary() {
        const elapsed = Math.round((Date.now() - this._startTime) / 100) / 10;
        // In watch mode, we want to count services that we started as part of this
        // run.
        const count = this._ran + this._servicesStarted;
        const s = count === 1 ? '' : 's';
        console.log(`✅ Ran ${count.toLocaleString()} script${s} and skipped ${this._skipped.toLocaleString()} in ${elapsed.toLocaleString()}s.`);
    }
    _reportOutputForFailingScript(script, cause) {
        const state = this._running.get(scriptReferenceToString(script));
        if (!state) {
            throw new Error(`Internal error: Got ${cause?.reason ? `${cause.reason} event` : 'leftover script'} for script without a start event. Events delivered out of order?
    Script with failure: ${scriptReferenceToString(script)}
    Known scripts: ${inspect([...this._running.keys()])}
`);
        }
        state.replayAndEmptyBuffer();
    }
    _getStatusLine() {
        switch (this._statusLineState) {
            case 'initial': {
                return 'Starting';
            }
            case 'analyzing': {
                return 'Analyzing';
            }
            case 'executing': {
                if (this._analysisInfo === undefined) {
                    return `??? Internal error: Analysis info missing ???`;
                }
                return this._getExecutionStatusLine(this._analysisInfo);
            }
            case 'done': {
                // No status line now.
                return nothing;
            }
            default: {
                const never = this._statusLineState;
                throw new Error(`Unknown status line state: ${JSON.stringify(never)}`);
            }
        }
    }
    _getExecutionStatusLine(analysisInfo) {
        const peekResult = this._running.peek()?.[1];
        let mostRecentScript = '';
        if (peekResult !== undefined) {
            mostRecentScript =
                ' ' + labelForScript(this._rootPackage, peekResult.scriptReference);
        }
        const done = this._finishedScriptsWithCommands.size;
        const total = analysisInfo.scriptsWithCommands.size;
        const percentDone = String(Math.round((done / total) * 100)).padStart(3, ' ') + '%';
        let servicesInfo = '';
        if (analysisInfo.hasServices) {
            const s = this._servicesRunning === 1 ? '' : 's';
            servicesInfo = ` [${this._servicesRunning.toLocaleString()} service${s}]`;
        }
        let failureInfo = '';
        if (this._scriptsWithAlreadyReportedErrors.size > 0) {
            failureInfo = ` [${this._scriptsWithAlreadyReportedErrors.size.toLocaleString()} failed]`;
        }
        return `${percentDone} [${done.toLocaleString()} / ${total}] [${this._running.size} running]${servicesInfo}${failureInfo}${mostRecentScript}`;
    }
    _markScriptAsFinished(script) {
        const key = scriptReferenceToString(script);
        // Almost always, a script that's finished will be one that we intended
        // to execute as part of this run. However there's one exception:
        // if a persistent service was started as part of the previous run, but then
        // we change the package.json files so that we no longer need to run it,
        // it will be shut down as part of _our_ run. In that case, we don't want
        // to count it towards our total.
        const isScriptOfInterest = 
        // Optimistically mark it as finished if we don't have analysis info yet.
        // We'll remove it later if we find out it's not actually a script we
        // care about.
        this._analysisInfo === undefined ||
            this._analysisInfo.scriptsWithCommands.has(key);
        if (isScriptOfInterest) {
            this._finishedScriptsWithCommands.add(key);
        }
    }
    _handleInfo(event) {
        if (DEBUG) {
            console.log(`info: ${event.detail} ${labelForScript(this._rootPackage, event.script)}`);
        }
        switch (event.detail) {
            case 'running': {
                const key = scriptReferenceToString(event.script);
                this._running.set(key, new ScriptState(event.script, false, this._analysisInfo?.rootScript === key));
                return this._getStatusLine();
            }
            case 'service-process-started': {
                // Services don't end, so we count this as having finished.
                this._servicesRunning++;
                this._servicesStarted++;
                const key = scriptReferenceToString(event.script);
                this._running.set(key, new ScriptState(event.script, true, this._analysisInfo?.rootScript === key));
                this._markScriptAsFinished(event.script);
                return this._getStatusLine();
            }
            case 'service-stopped':
                this._servicesRunning--;
                this._running.delete(scriptReferenceToString(event.script));
                return this._getStatusLine();
            case 'service-ready':
            case 'watch-run-start':
            case 'watch-run-end':
                return noChange;
            case 'locked': {
                // the script is blocked on starting because something else is
                // using a shared resource
                return noChange;
            }
            case 'output-modified': {
                // the script is being run because its output files have changed
                // log nothing
                return noChange;
            }
            case 'cache-info': {
                // chatty event, e.g. transitory GitHub API errors.
                // definitely don't log anything here
                return noChange;
            }
            case 'analysis-started': {
                this._statusLineState = 'analyzing';
                return this._getStatusLine();
            }
            case 'analysis-completed': {
                if (!event.rootScriptConfig) {
                    // will report the error in printSummary
                    this._statusLineState = 'done';
                    return nothing;
                }
                else {
                    this._statusLineState = 'executing';
                    this._analysisInfo = this._countScriptsWithCommands(event.rootScriptConfig);
                    for (const finished of this._finishedScriptsWithCommands) {
                        if (!this._analysisInfo.scriptsWithCommands.has(finished)) {
                            this._finishedScriptsWithCommands.delete(finished);
                        }
                    }
                }
                return this._getStatusLine();
            }
            default: {
                const never = event;
                throw new Error(`Unknown info event: ${JSON.stringify(never)}`);
            }
        }
    }
    _handleSuccess(event) {
        if (DEBUG) {
            console.log(`success: ${event.reason} ${labelForScript(this._rootPackage, event.script)}`);
        }
        switch (event.reason) {
            case 'cached': {
                this._markScriptAsFinished(event.script);
                this._skipped++;
                return this._getStatusLine();
            }
            case 'fresh': {
                this._markScriptAsFinished(event.script);
                this._skipped++;
                return this._getStatusLine();
            }
            case 'exit-zero': {
                this._running.delete(scriptReferenceToString(event.script));
                this._markScriptAsFinished(event.script);
                this._ran++;
                return this._getStatusLine();
            }
            case 'no-command': {
                return noChange;
            }
            default: {
                const never = event;
                throw new Error(`Unknown success event: ${JSON.stringify(never)}`);
            }
        }
    }
    _handleFailure(event) {
        if (DEBUG) {
            console.log(`failure: ${event.reason} ${labelForScript(this._rootPackage, event.script)}`);
        }
        this._encounteredFailures = true;
        {
            const env_1 = { stack: [], error: void 0, hasError: false };
            try {
                // eslint-disable-next-line @typescript-eslint/no-unused-vars
                const _pause = __addDisposableResource(env_1, this._statusLineWriter.clearUntilDisposed(), false);
                this._reportFailure(event);
            }
            catch (e_1) {
                env_1.error = e_1;
                env_1.hasError = true;
            }
            finally {
                __disposeResources(env_1);
            }
        }
        return noChange;
    }
    _reportFailure(failure) {
        const key = scriptReferenceToString(failure.script);
        if (this._scriptsWithAlreadyReportedErrors.has(key)) {
            return;
        }
        this._scriptsWithAlreadyReportedErrors.add(key);
        const label = labelForScript(this._rootPackage, failure.script);
        switch (failure.reason) {
            case 'exit-non-zero':
            case 'signal':
            case 'killed': {
                let message;
                if (failure.reason === 'exit-non-zero') {
                    message = `exited with exit code ${failure.status}`;
                }
                else if (failure.reason === 'signal') {
                    message = `was killed by signal ${failure.signal}`;
                }
                else {
                    message = `killed`;
                }
                const scriptHadOutput = this._scriptHadOutput(failure.script);
                const trailer = scriptHadOutput ? ' Output:\n' : '';
                process.stderr.write(`\n❌ [${label}] ${message}.${trailer}\n`);
                if (scriptHadOutput) {
                    this._reportOutputForFailingScript(failure.script, failure);
                }
                this._finishedScriptsWithCommands.add(key);
                this._running.delete(key);
                return this._getStatusLine();
            }
            case 'start-cancelled':
            case 'aborted': {
                // These events aren't very useful to log, because they are downstream
                // of failures that already get reported elsewhere.
                this._running.delete(scriptReferenceToString(failure.script));
                return this._getStatusLine();
            }
            case 'dependency-service-exited-unexpectedly': {
                // Also logged elswhere.
                break;
            }
            case 'service-exited-unexpectedly':
            case 'cycle':
            case 'dependency-invalid':
            case 'dependency-on-missing-package-json':
            case 'dependency-on-missing-script':
            case 'duplicate-dependency':
            case 'failed-previous-watch-iteration':
            case 'invalid-config-syntax':
            case 'invalid-json-syntax':
            case 'invalid-usage':
            case 'launched-incorrectly':
            case 'missing-package-json':
            case 'no-scripts-in-package-json':
            case 'script-not-found':
            case 'script-not-wireit':
            case 'spawn-error':
            case 'unknown-error-thrown':
            case 'wireit-config-but-no-script':
                // The default log for these is good.
                this._defaultLogger.log(failure);
                break;
            default: {
                const never = failure;
                throw new Error(`Unknown failure event: ${JSON.stringify(never)}`);
            }
        }
    }
    _scriptHadOutput(script) {
        const state = this._running.get(scriptReferenceToString(script));
        if (!state) {
            throw new Error(`Internal error: could not find state for failing script. Events delivered out of order?
        Script with output: ${labelForScript(this._rootPackage, script)}
        ${this._running.size.toLocaleString()} known running scripts: ${inspect([...this._running.keys()])}`);
        }
        return state.hasBufferedOutput;
    }
    _handleOutput(event) {
        const key = scriptReferenceToString(event.script);
        const state = this._running.get(key);
        if (!state) {
            throw new Error(`Internal error: Got output event for unknown script. Events delivered out of order?
        Script with output: ${labelForScript(this._rootPackage, event.script)}
        ${this._running.size.toLocaleString()} known running scripts: ${inspect([...this._running.keys()])}`);
        }
        if (state.service) {
            const env_2 = { stack: [], error: void 0, hasError: false };
            try {
                // Pause the status line while we print this real quick, but then resume
                // it.
                // eslint-disable-next-line @typescript-eslint/no-unused-vars
                const _pause = __addDisposableResource(env_2, this._statusLineWriter.clearUntilDisposed(), false);
                if (event.stream === 'stdout') {
                    process.stdout.write(event.data);
                }
                else {
                    process.stderr.write(event.data);
                }
                return noChange;
            }
            catch (e_2) {
                env_2.error = e_2;
                env_2.hasError = true;
            }
            finally {
                __disposeResources(env_2);
            }
        }
        if (state.isRootScript) {
            this._statusLineState = 'done';
            // Unlike for a service, this is a terminal state, instead of pausing the
            // status line, we stop it completely, because for the rest of the run
            // we're just going to be  printing the root script's output.
            this._statusLineWriter.clearAndStopRendering();
            if (event.stream === 'stdout') {
                process.stdout.write(event.data);
            }
            else {
                process.stderr.write(event.data);
            }
            return nothing;
        }
        // Buffer everything else so that we can print it
        // (possibly a second time) in case of failure.
        state.bufferOutput(event);
        return noChange;
    }
    _countScriptsWithCommands(rootScript) {
        const scriptsWithCommands = new Set();
        let hasServices = false;
        const seen = new Set([scriptReferenceToString(rootScript)]);
        const toVisit = [rootScript];
        while (toVisit.length > 0) {
            const script = toVisit.pop();
            if (script.service) {
                hasServices = true;
            }
            if (script.command !== undefined) {
                // We only want to count scripts that actually run, rather than
                // just holding dependencies.
                scriptsWithCommands.add(scriptReferenceToString(script));
            }
            for (const dependency of script.dependencies.values()) {
                const key = scriptReferenceToString(dependency.config);
                if (seen.has(key)) {
                    continue;
                }
                seen.add(key);
                toVisit.push(dependency.config);
            }
        }
        return {
            rootScript: scriptReferenceToString(rootScript),
            scriptsWithCommands,
            hasServices,
        };
    }
}
//# sourceMappingURL=run-tracker.js.map