/**
 * @license
 * Copyright 2023 Google LLC
 * SPDX-License-Identifier: Apache-2.0
 */
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _Semaphore_remaining, _Semaphore_waiting;
import * as fs from 'fs/promises';
import { createReadStream as rawCreateReadStream, createWriteStream as rawCreateWriteStream, } from 'fs';
import { Deferred } from './deferred.js';
import './dispose.js';
export { constants } from 'fs';
export class Semaphore {
    constructor(numSlots) {
        _Semaphore_remaining.set(this, void 0);
        _Semaphore_waiting.set(this, []);
        if (numSlots <= 0) {
            throw new Error(`numSlots must be positive, got ${numSlots}`);
        }
        __classPrivateFieldSet(this, _Semaphore_remaining, numSlots, "f");
    }
    async reserve() {
        var _a;
        while (__classPrivateFieldGet(this, _Semaphore_remaining, "f") === 0) {
            const deferred = new Deferred();
            __classPrivateFieldGet(this, _Semaphore_waiting, "f").push(deferred);
            await deferred.promise;
        }
        __classPrivateFieldSet(this, _Semaphore_remaining, (_a = __classPrivateFieldGet(this, _Semaphore_remaining, "f"), _a--, _a), "f");
        let disposed = false;
        return {
            [Symbol.dispose]: () => {
                var _a;
                if (disposed) {
                    return;
                }
                disposed = true;
                __classPrivateFieldSet(this, _Semaphore_remaining, (_a = __classPrivateFieldGet(this, _Semaphore_remaining, "f"), _a++, _a), "f");
                if (__classPrivateFieldGet(this, _Semaphore_waiting, "f").length > 0) {
                    __classPrivateFieldGet(this, _Semaphore_waiting, "f").pop()?.resolve();
                }
            },
        };
    }
}
_Semaphore_remaining = new WeakMap(), _Semaphore_waiting = new WeakMap();
export const fileBudget = (() => {
    let maxOpenFiles = Number(process.env['WIREIT_MAX_OPEN_FILES']);
    if (isNaN(maxOpenFiles)) {
        // This is tricky to get right. There's no simple cross-platform way to
        // determine what our current limits are. Windows it's 512, on macOS it
        // defaults to 256, and on Linux it varies a lot.
        // 200 gives us a bit of headroom for other things that might be using
        // file descriptors in our process, like node internals.
        maxOpenFiles = 200;
    }
    return new Semaphore(maxOpenFiles);
})();
export async function mkdir(path, options) {
    const reservation = await fileBudget.reserve();
    try {
        return fs.mkdir(path, options);
    }
    finally {
        reservation[Symbol.dispose]();
    }
}
export async function mkdtemp(path) {
    const reservation = await fileBudget.reserve();
    try {
        return fs.mkdtemp(path);
    }
    finally {
        reservation[Symbol.dispose]();
    }
}
export async function writeFile(path, contents, encoding) {
    const reservation = await fileBudget.reserve();
    try {
        return fs.writeFile(path, contents, encoding);
    }
    finally {
        reservation[Symbol.dispose]();
    }
}
export async function readFile(path, encoding) {
    const reservation = await fileBudget.reserve();
    try {
        return fs.readFile(path, encoding);
    }
    finally {
        reservation[Symbol.dispose]();
    }
}
export async function rm(path, options) {
    const reservation = await fileBudget.reserve();
    try {
        return fs.rm(path, options);
    }
    finally {
        reservation[Symbol.dispose]();
    }
}
export async function lstat(path) {
    const reservation = await fileBudget.reserve();
    try {
        return fs.lstat(path);
    }
    finally {
        reservation[Symbol.dispose]();
    }
}
export async function stat(path) {
    const reservation = await fileBudget.reserve();
    try {
        return fs.stat(path);
    }
    finally {
        reservation[Symbol.dispose]();
    }
}
export async function access(path) {
    const reservation = await fileBudget.reserve();
    try {
        return fs.access(path);
    }
    finally {
        reservation[Symbol.dispose]();
    }
}
export async function createReadStream(path, options) {
    const reservation = await fileBudget.reserve();
    const stream = rawCreateReadStream(path, options);
    stream.on('close', () => reservation[Symbol.dispose]());
    return stream;
}
export async function createWriteStream(path) {
    const reservation = await fileBudget.reserve();
    const stream = rawCreateWriteStream(path);
    stream.on('close', () => reservation[Symbol.dispose]());
    return stream;
}
export async function copyFile(src, dest, flags) {
    const reservation = await fileBudget.reserve();
    try {
        return await fs.copyFile(src, dest, flags);
    }
    finally {
        reservation[Symbol.dispose]();
    }
}
export async function readlink(path, options) {
    const reservation = await fileBudget.reserve();
    try {
        return await fs.readlink(path, options);
    }
    finally {
        reservation[Symbol.dispose]();
    }
}
export async function symlink(target, path, type) {
    const reservation = await fileBudget.reserve();
    try {
        return await fs.symlink(target, path, type);
    }
    finally {
        reservation[Symbol.dispose]();
    }
}
export async function unlink(target) {
    const reservation = await fileBudget.reserve();
    try {
        return await fs.unlink(target);
    }
    finally {
        reservation[Symbol.dispose]();
    }
}
export async function rmdir(target) {
    const reservation = await fileBudget.reserve();
    try {
        return await fs.rmdir(target);
    }
    finally {
        reservation[Symbol.dispose]();
    }
}
//# sourceMappingURL=fs.js.map